# Generated file
# - do not overwrite
# - do not include in git
from collections import namedtuple

BuildInfo = namedtuple(
    'BuildInfo',
    ['version', 'built_at', 'sha', 'branch', 'tag'],
)

BUILD_INFO = BuildInfo(
    version='v4.28.4~107dd95',
    built_at='2026-03-25 16:03:18Z',
    sha='107dd9598b34665da78dd5b289fcb27488b796f0',
    branch='HEAD',
    tag='v4.28.4',
)
